-- Start Code --

--[[
KNX Thermostat Driver. Allows homeowner to control KNX Thermostat devices. Used with KNX Master driver.
Copyright 2017 Control4 Corporation. All Rights Reserved.
]]

--[[

TODO: Figure out why my room control is lying to me about Read Request on 3/4/0... Doesn't match what I sent it, doesn't seem to change unless I change the mode on the room control itself. Seems wrong. Causing this: "On restart, mode not successfully set from status value from thermostat..."

TODO: Lukas Polivka Critiques:

 * Apply to all thermostats (while copying schedules) has no effect on other KNX Thermostat drivers in the project. (Is that also related to the reserved preset field IDs?)

* TODO: 'Mode DPT' / 'Ventilation Mode DPT'

--]]


KNX_BINDING = 1
PROXY_BINDING = 5001

PROP_VISIBLE = 0
PROP_HIDDEN = 1

F_RESOLUTIONS = {}
F_RESOLUTIONS["1"] = "1"
F_RESOLUTIONS[".5"] = ".5"
F_RESOLUTIONS[".1"] = ".2"


-- Test function, used to issue a read request to KNX bus on specified Group Address, to test incoming KNX data...
-- Example: RequestStatus(Properties["Temperature Status Address"])
--   Causes Temperature to be requested and then sent to this driver...
function RequestStatus(ga)
  MySendToProxy(KNX_BINDING, "REQUEST_STATUS", {GA = ga})
end


function RequestStatusAll()
  for k,v in pairs(Properties) do if (k:find("Status") and (trim(v) ~= "")) then RequestStatus(v) end end
end


function OnPropertyChanged(strProperty)
  local prop = Properties[strProperty]
  if (strProperty == "Debug Mode") then
    if ((gDbgTimer or 0) > 0) then gDbgTimer = C4:KillTimer(gDbgTimer) end
    gDbgPrint, gDbgLog = (prop:find("Print") ~= nil), (prop:find("Log") ~= nil)
    if (prop == "Off") then return end
    gDbgTimer = C4:AddTimer(8*60, "MINUTES")
    dbg("Started Debug timer for 8 hours")
    return
  end
  if (strProperty == "Ventilation Mode List") then
    if (Properties["Ventilation Control"] ~= "Off") then
      gFanModeList, _gFanModeList = ParseFanModeList()
    end
  end
  if (strProperty == "Mode List") then
    gModeList, _gModeList = ParseModeList()
    SetupPresetScheduling() -- Setup Preset fields.
    return
  end
  if (strProperty == "Preset Scheduling") then
    SetupPresetScheduling()
    return
  end
  if (strProperty == "Ventilation Control") then
    if (prop == "Off") then
      for k,v in pairs({"Ventilation Set Address","Ventilation Status Address","Ventilation Mode List","Ventilation Mode DPT"}) do C4:SetPropertyAttribs(v, PROP_HIDDEN) end
      MySendToProxy(PROXY_BINDING, "ALLOWED_FAN_MODES_CHANGED", {MODES = " "})
    else
      for k,v in pairs({"Ventilation Set Address","Ventilation Status Address","Ventilation Mode List","Ventilation Mode DPT"}) do C4:SetPropertyAttribs(v, PROP_VISIBLE) end
      OnPropertyChanged("Ventilation Mode List")
    end
    return
  end
  if (strProperty == "Mode Temperature Ranges") then
    gModeRange = ParseModeRanges()
    return
  end
  if (strProperty == "Hold Modes") then
    if (prop == "Off") then
      MySendToProxy(PROXY_BINDING, "ALLOWED_HOLD_MODES_CHANGED", {MODES = " "})
    else
      MySendToProxy(PROXY_BINDING, "ALLOWED_HOLD_MODES_CHANGED", {MODES = "Off,2 Hours,Until Next,Permanent"})
    end
    return
  end

  if (strProperty == "Setpoint Resolution") then
    MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {SINGLE_SETPOINT_RESOLUTION_C = prop})
    MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {SINGLE_SETPOINT_RESOLUTION_F = F_RESOLUTIONS[prop]})
    return
  end
  if (strProperty == "Temperature Resolution") then
    MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {TEMPERATURE_RESOLUTION_C = prop})
    MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {TEMPERATURE_RESOLUTION_F = F_RESOLUTIONS[prop]})
    return
  end

  OneShotTimer.Add(5, "SECONDS", AddGroupItems, "AGI")
end


function OnDriverDestroyed()
  gDbgTimer = C4:KillTimer(gDbgTimer or 0)
end


function dbg(strDebugText)
  if (gDbgPrint) then print (strDebugText) end
  if (gDbgLog) then C4:ErrorLog("KNX Thermostat: " .. strDebugText) end
end


function formatParams(tParams)
  tParams = tParams or {}
  local out = {}
  for k,v in pairs(tParams) do
    if (type(v) == "string") then
      table.insert(out, k .. ': "' .. tostring(v):sub(1, 1000) .. '"')
    else
      table.insert(out, k .. ": " .. tostring(v):sub(1, 1000))
    end
  end
  return table.concat(out, ", ")
end


function OnTimerExpired(idTimer)
  if (idTimer == gDbgTimer) then
    dbg("Turning Debug Mode Off (timer expired)")
    C4:UpdateProperty("Debug Mode", "Off")
    OnPropertyChanged("Debug Mode")
    gDbgTimer = C4:KillTimer(gDbgTimer)
  end

  -- Kill any 'stray' timers (should not be needed)
  C4:KillTimer(idTimer)
end


function trim(s)
  s = s or ""
  return (s:gsub("^%s*(.-)%s*$", "%1")) -- from PiL2 20.4
end


-- Customized version of tonumber... Should work regardless of locale...
function tonumber_loc(str, base)
  if (type(str) ~= "string") then str = tostring(str) end

  local s = str:gsub(",", ".") -- Assume US Locale decimal separator
  local num = tonumber(s, base)
  if (num == nil) then
    s = str:gsub("%.", ",") -- Non-US Locale decimal separator
    num = tonumber(s, base)
  end
  return num
end


function SendToKNX(tParams)
  tParams = tParams or {}
  -- Use 'canonical' Network Driver name for DPT and GA...
  tParams.DATA_POINT_TYPE = tParams.DATA_POINT_TYPE or tParams.DPT
  tParams.DPT = nil
  tParams.GROUP_ADDRESS = tParams.GROUP_ADDRESS or tParams.GA
  tParams.GA = nil

  MySendToProxy(KNX_BINDING, "SEND_TO_KNX", tParams)
end


-- Set Setpoint example...
-- SendToKNX({ DATA_POINT_TYPE = "DPT_9", GROUP_ADDRESS = Properties["Setpoint Set Address"], VALUE = "18.5"})
-- SendToKNX({ DPT = "DPT_9", GA = Properties["Setpoint Set Address"], VALUE = "18.5"})

----------------------- THERMOSTAT HELPER FUNCTIONS -------------------------
FEEDBACK_ITEMS = {}
FEEDBACK_ITEMS["Temperature Status Address"] = "DPT_9"
FEEDBACK_ITEMS["Setpoint Status Address"] = "DPT_9"
FEEDBACK_ITEMS["Mode Status Address"] = "DPT_5"
FEEDBACK_ITEMS["Ventilation Status Address"] = "DPT_5"
FEEDBACK_ITEMS["Heating Status Address"] = "DPT_1"
FEEDBACK_ITEMS["Cooling Status Address"] = "DPT_1"


function OnBindingChanged(idBinding, strClass, bConnected)
  if (strClass == "KNX_DEVICE") and (bConnected) then
    dbg("KNX Binding Changed. Sending Group Addresses...")
    OneShotTimer.Add(10, "SECONDS", AddGroupItems, "AGI")
  end
end


function AddGroupItem(propName, dpt)
  local prop = trim(Properties[propName])
  if (prop == "") then return end

  MySendToProxy(KNX_BINDING, "ADD_GROUP_ITEM", {GROUP_ADDRESS = prop, DEVICE_ID = C4:GetDeviceID(), PROPERTY = propName, DATA_POINT_TYPE = dpt})
end


function AddGroupItems()
  RemoveGroupItems()

  -- Update DPT for Heating/Cooling Status...
  local _, _, dpt = (Properties["Heating Status DPT"] or "DPT_1 "):find("(DPT_%d) .*")
  FEEDBACK_ITEMS["Heating Status Address"] = dpt

  _, _, dpt = (Properties["Cooling Status DPT"] or "DPT_1 "):find("(DPT_%d) .*")
  FEEDBACK_ITEMS["Cooling Status Address"] = dpt

  local _, _, dpt = (Properties["Mode DPT"] or "DPT_5 "):find("(DPT_%d) .*")
  FEEDBACK_ITEMS["Mode Status Address"] = dpt

  local _, _, dpt = (Properties["Ventilation Mode DPT"] or "DPT_5 "):find("(DPT_%d) .*")
  FEEDBACK_ITEMS["Ventilation Status Address"] = dpt

  for item, dpt in pairs(FEEDBACK_ITEMS) do
    AddGroupItem(item, dpt)
  end
end


function RemoveGroupItems()
  local tParams = {DEVICE_ID = C4:GetDeviceID()}
  MySendToProxy(KNX_BINDING, "CLEAR_GROUP_ITEMS", tParams)
end


function FtoC(valueF)
  -- Verify sane input...
  valueF = tonumber_loc(valueF) or 0

  -- °C = (°F - 32) * 5/9
  local valueC = ((valueF - 32) * 5 / 9)

  -- No need to round...
  return valueC
end

function CtoF(valueC)
  -- Verify sane input...
  valueC = tonumber_loc(valueC) or 0

  -- °F = (°C * 9 / 5) + 32
  local valueF = (valueC * 9 / 5) + 32

  -- No need to round...
  return valueF
end


FROM_BINDING = {}
FROM_BINDING[KNX_BINDING] = "<---- KNX "
FROM_BINDING[PROXY_BINDING] = "<---- Proxy"


-- TODO: Need to preserve a way to *not* write the .po files, in case installer needs to do manual .po rewrite...
-- i.e. if the 'original' already has _Heat, _Cool, or _*, we don't rewrite .po file with those values...
-- Don't rewrite .po files if no .po values found in calling function...

LANGUAGES_DIR = "/mnt/internal/c4z/knx_thermostat/www/languages/"

-- Rewrites .po files in www/languages/*.po on controller...
function RewritePOFiles(po_rewrites)

  local out, tins = {}, table.insert
  for _,item in pairs(po_rewrites) do 
    tins(out, "msgid \"" .. item.msgid .. "\"")
    tins(out, "msgstr \"" .. item.msgstr .. "\"\n")
  end

  C4:FileSetDir(LANGUAGES_DIR)
  local filelist = C4:FileList()
  local filecount = 0
  for k,v in pairs(filelist) do filecount = filecount + 1 end
  dbg("Rewriting " .. filecount .. " .po files in " .. LANGUAGES_DIR)

  for _,v in pairs(filelist) do
    -- Check that it's a .po file...
    if (v:find("\.po")) then
      local fh = C4:FileOpen(v)
      C4:FileSetPos(fh, 0)
      local content = C4:FileRead(fh, 50000)
      C4:FileClose(fh)

      local _, fstart = content:find("%[AUTO_GENERATED_START%]\n")
      local fend = content:find("#%[AUTO_GENERATED_END%]")
      local newcontent = content:sub(0, fstart) .. "\n" .. table.concat(out, "\n") .. "\n" .. content:sub(fend)

      local fh = C4:FileOpen(v)
      C4:FileSetPos(fh, 0)
      C4:FileWriteString(fh, newcontent)
      C4:FileClose(fh)
    end
  end

  C4:FileSetDir("")
end


function ParseModeList()
  local ml = Properties["Mode List"] .. ","
  local ModeList = {}
  local _ModeList = {}
  local allowed = {}
  local po_rewrites = {}

  for item in ml:gfind("(.-),") do
    local _,_,modename,id = item:find("(.+)=(.+)")
    modename,id = trim(modename), trim(id)
    local orig = modename

    -- Automatically insert _Heat if no 'mode-type' in mode name...
    local first, last, capture = modename:find('%[(.+)%]')
    if (first) then
      orig = modename:sub(0,first-1) .. modename:sub(last+1)
      modename = orig .. "_" .. capture
    else
      modename = modename .. "_Heat"
    end

    --dbg("msgstr: " .. orig .. " msgid: " .. modename) -- msgid "Comfort_Off" -- msgstr "Comfort"
    table.insert(po_rewrites, {msgstr = orig, msgid = modename})

    ModeList[modename] = id
    _ModeList[id] = modename
    table.insert(allowed, modename)
  end

  MySendToProxy(PROXY_BINDING, "ALLOWED_HVAC_MODES_CHANGED", {MODES = table.concat(allowed, ",")})
  RewritePOFiles(po_rewrites)

  return ModeList, _ModeList
end


function ParseFanModeList()
  local ml = Properties["Ventilation Mode List"] .. ","
  local ModeList = {}
  local _ModeList = {}
  local allowed = {}

  for item in ml:gfind("(.-),") do
    local _,_,modename,id = item:find("(.+)=(.+)")
    modename,id = trim(modename), trim(id)
    ModeList[modename] = id
    _ModeList[id] = modename
    table.insert(allowed, modename)
  end

  MySendToProxy(PROXY_BINDING, "ALLOWED_FAN_MODES_CHANGED", {MODES = table.concat(allowed, ",")})

  return ModeList, _ModeList
end


function ParseModeRanges()
  local ml = Properties["Mode Temperature Ranges"] .. ","
  local ModeRange = {}

  for item in ml:gfind("(.-),") do
    local _,_,num,range = item:find("(.+)=(.+)")
    num, range = trim(num), trim(range)
    local _,_,min,max = range:find("(.+)-(.+)")
    min, max = trim(min), trim(max)
    ModeRange[num] = {}
    ModeRange[num].min = min
    ModeRange[num].max = max
  end
  return ModeRange
end


function SetRanges(mode)
  local modenum = gModeList[mode] or {}
  gModeRange = ParseModeRanges()
  local modes = gModeRange[modenum] or {}
  if (modes.min ~= nil ) then
    MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {SINGLE_SETPOINT_MIN_C = modes.min, SINGLE_SETPOINT_MIN_F = tostring(CtoF(modes.min))})
    gCurrentMin = tonumber_loc(modes.min) or -255
  end
  if (modes.max ~= nil ) then
    MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {SINGLE_SETPOINT_MAX_C = modes.max, SINGLE_SETPOINT_MAX_F = tostring(CtoF(modes.max))})
    gCurrentMax = tonumber_loc(modes.max) or -255
  end
end


function DeletePresets()
  -- Get Bound Proxy's Device ID / Name.
  local proxyid, name = next(C4:GetBoundConsumerDevices(C4:GetDeviceID(), PROXY_BINDING))
  local presetxml = C4:SendUIRequest(proxyid, "GET_PRESETS", {})

  -- For each preset, delete it.
  local presets = XMLHelper.XMLToTable(presetxml)
  for i, v in ipairs(presets[1]) do
    C4:SendToDevice(proxyid, "PRESET_DELETE", {NAME = v.xarg.name})
  end
end


function AddModePresets()
  -- Get Bound Proxy's Device ID / Name.
  local proxyid, name = next(C4:GetBoundConsumerDevices(C4:GetDeviceID(), PROXY_BINDING))

  for k,v in pairs(gModeList) do
    local preset_fields = '<preset_fields><field id="hvac_mode" value="' .. k .. '" /></preset_fields>'
    C4:SendToDevice(proxyid, "PRESET_ADD", {NAME = k:match("(.*)_.*"), PRESET_FIELDS = preset_fields})
  end
end


function GetRangeMinMax()
  -- Find min/max of all min/max values...
  local min, max = 2000, -2000
  for k,v in pairs(gModeRange) do
    if (tonumber_loc(v.min) < min) then min = tonumber_loc(v.min) end
    if (tonumber_loc(v.max) > max) then max = tonumber_loc(v.max) end
  end
  return min, max
end


function SetPresetFields(preset_fields)
  MySendToProxy(PROXY_BINDING, "PRESET_FIELDS_CHANGED", {XML = preset_fields})
  PersistData.PresetFields = preset_fields
  C4:InvalidateState()
end


function SetupPresetScheduling()
  local preset_scheduling = Properties["Preset Scheduling"]

  gLastScheduledPreset = ""
  PersistData = PersistData or {}
  DeletePresets()

  if (preset_scheduling == "Off") then
    MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {CAN_PRESET = false})
    MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {CAN_PRESET_SCHEDULE = false})
    SetPresetFields("")
    return
  end

  MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {CAN_PRESET = true})
  MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {CAN_PRESET_SCHEDULE = true})
  gModeRange = ParseModeRanges()
  gModeList, _gModeList = ParseModeList()
  local min, max = GetRangeMinMax()
  local res = Properties["Setpoint Resolution"] or ".5"

  if (preset_scheduling == "Mode Only") then
    local preset_fields = '<preset_fields>'
    preset_fields = preset_fields .. '<field id="hvac_mode" type="list" label="Mode"><list>'
    for k,v in pairs(gModeList) do
      preset_fields = preset_fields .. '<item text="' .. k .. '" value="' .. k .. '" />'
    end
    preset_fields = preset_fields .. '</list></field>' 
    preset_fields = preset_fields .. '</preset_fields>'
    AddModePresets()
    SetPresetFields(preset_fields)
  elseif (preset_scheduling == "Mode+Temp") then
    local preset_fields = '<preset_fields>'
    preset_fields = preset_fields .. '<field id="hvac_mode" type="list" label="Mode"><list>'
    for k,v in pairs(gModeList) do
      preset_fields = preset_fields .. '<item text="' .. k .. '" value="' .. k .. '" />'
    end
    preset_fields = preset_fields .. '</list></field>' 
    preset_fields = preset_fields .. '<field id="single_setpoint_c" type="number" label="Setpoint" min="' .. min .. '" max="' .. max .. '" res="' .. res .. '" />'
    preset_fields = preset_fields .. '</preset_fields>'
    SetPresetFields(preset_fields)
  elseif (preset_scheduling == "Temperature Only") then
    local preset_fields = '<preset_fields>'
    preset_fields = preset_fields .. '<field id="single_setpoint_c" type="number" label="Setpoint" min="' .. min .. '" max="' .. max .. '" res="' .. res .. '" />'
    preset_fields = preset_fields .. '</preset_fields>'
    SetPresetFields(preset_fields)
  end
end


function RestorePresetScheduling()
  local preset_scheduling = Properties["Preset Scheduling"]

  if (preset_scheduling == "Off") then
    DeletePresets()
    MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {CAN_PRESET = false})
    MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {CAN_PRESET_SCHEDULE = false})
    PersistData.PresetFields = ""
    return
  else
    MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {CAN_PRESET = true})
    MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {CAN_PRESET_SCHEDULE = true})
    PersistData = PersistData or {}
    local preset_fields = PersistData.PresetFields or ""
    if (preset_fields ~= "") then
      MySendToProxy(PROXY_BINDING, "PRESET_FIELDS_CHANGED", {XML = preset_fields})
    end
  end
end


function ReceivedFromProxy(idBinding, strCommand, tParams)
  strCommand = strCommand or ""
  tParams = tParams or {}

  if (not IGNORE_MESSAGES[strCommand]) then
    dbg("[" .. FROM_BINDING[idBinding] .. "]: " .. strCommand .. " {" .. formatParams(tParams) .. "}")
  end

  -- TODO: Handle Proxy EXTRAS commands...
  PersistData = PersistData or {}

  if (strCommand == "SET_SCALE") then
    local scale = tParams["SCALE"] or ""
    if (scale == "") then return end
    gMyScale = scale
    PersistData.Scale = scale
    MySendToProxy(idBinding, "SCALE_CHANGED", {SCALE = gMyScale})
  end

  -- Do a conversion to the 'KELVIN' param if not present in params...
  local kelvin = tParams["KELVIN"] or 0
  local setpoint = tParams["SETPOINT"] or 0
  local scale = tParams["SCALE"] or "F"

  -- If we have a Scale and Setpoint, but no Kelvin param, make one...
  if (kelvin == 0) and (setpoint ~= 0) and (scale ~= "") then
    if (scale == "F") then setpoint = FtoC(setpoint)
      kelvin = (setpoint + 273.15) * 10
      tParams["KELVIN"] = kelvin
    end
  end

  if (strCommand == "SET_SETPOINT_SINGLE") then

    local value = tonumber_loc(tParams["CELSIUS"]) or -255
    if (value == -255) then return end

    local isPreset = tParams["PRESET"] or false

    -- If we have min/max for current mode, bracket temp within min/max
    gCurrentMin = gCurrentMin or -255
    gCurrentMax = gCurrentMax or -255
    if (gCurrentMin > -255) and (gCurrentMax > -255) then
      if (value < gCurrentMin) then dbg("Setpoint [" .. value .. "] increased into range --> [" .. gCurrentMin .. "]") value = gCurrentMin end
      if (value > gCurrentMax) then dbg("Setpoint [" .. value .. "] decreased into range --> [" .. gCurrentMax .. "]") value = gCurrentMax end
    end

    -- TODO: What if this value is in Fahrenheit on Room Control / Thermostat device???
    SendToKNX({DATA_POINT_TYPE = "DPT_9", GROUP_ADDRESS = Properties["Setpoint Set Address"], VALUE = tostring(value)})

    if (not isPreset) then
      -- Only set Hold Until Next if presets are done with Temperature (not for Off or Mode Only)
      local psmode = Properties["Preset Scheduling"] or ""
      if (psmode == "Mode+Temp") or (psmode == "Temperature Only") then
        ReceivedFromProxy(PROXY_BINDING, "SET_MODE_HOLD", {MODE = "Until Next"})
        MySendToProxy(PROXY_BINDING, "PRESET_CHANGED", {NAME = ""})
      end
    end

    -- Change UI if no response from KNX for 5 seconds...
    OneShotTimer.Add(5, "SECONDS", function() MySendToProxy(PROXY_BINDING, "SINGLE_SETPOINT_CHANGED", {SCALE = "C", SETPOINT = tostring(value)}) end, "SSC")
  end

  if (strCommand == "SET_MODE_HVAC") then
    local mode = tParams["MODE"] or ""
    if (mode == "") then return end
    local isPreset = tParams["PRESET"] or false

    local _, _, dpt = (Properties["Mode DPT"] or "DPT_5 "):find("(DPT_%d) .*")
    SendToKNX({DPT = dpt, GA = Properties["Mode Set Address"], VALUE = gModeList[mode]})
    SetRanges(mode)

    -- Change UI if no response from KNX for 5 seconds...
    OneShotTimer.Add(5, "SECONDS", function() MySendToProxy(PROXY_BINDING, "HVAC_MODE_CHANGED", {MODE = mode}) end, "HMC")

    if (not isPreset) then
      -- Only set Hold Until Next if presets are done with Mode (not for Off or Temperature Only)
      local psmode = Properties["Preset Scheduling"] or ""
      if (psmode == "Mode+Temp") or (psmode == "Mode Only") then
        ReceivedFromProxy(PROXY_BINDING, "SET_MODE_HOLD", {MODE = "Until Next"})
        MySendToProxy(PROXY_BINDING, "PRESET_CHANGED", {NAME = ""})
      end
    end
  end

  if (strCommand == "SET_MODE_FAN") then
    local mode = tParams["MODE"] or ""
    if (mode == "") then return end

    local _, _, dpt = (Properties["Ventilation Mode DPT"] or "DPT_5 "):find("(DPT_%d) .*")
    SendToKNX({DPT = dpt, GA = Properties["Ventilation Set Address"], VALUE = gFanModeList[mode]})

    -- Change UI if no response from KNX for 5 seconds...
    OneShotTimer.Add(5, "SECONDS", function() MySendToProxy(PROXY_BINDING, "FAN_MODE_CHANGED", {MODE = mode}) end, "FMC")
  end

  if (strCommand == "SET_MODE_HOLD") then
    gPrevMode = gHoldMode
    gHoldMode = tParams["MODE"] or "Off"
    if (gHoldMode == gPrevMode) then return end

    MySendToProxy(PROXY_BINDING, "HOLD_MODE_CHANGED", tParams)

    -- When Hold Mode is turned off, restore 'current' preset...
    if (gHoldMode == "Off") then
      if ((gLastScheduledPreset or "") ~= "") then
        dbg("Hold Mode Off.  Restoring Preset [" .. gLastScheduledPreset .. "]")
        ReceivedFromProxy(PROXY_BINDING, "SET_PRESET", {NAME = gLastScheduledPreset})
      end
    end

    OneShotTimer.Kill("HOLDMODE")

    -- Set a Hold Mode timer for 2 hours mode only... Everything else, no timer needed
    if (gHoldMode == "2 Hours") then
      OneShotTimer.Add(2, "HOURS", 
        function() 
          ReceivedFromProxy(PROXY_BINDING, "SET_MODE_HOLD", {MODE = "Off"})
        end
      , "HOLDMODE")
    end
    return
  end

  if (strCommand == "SET_EVENTS") then
    -- TODO: Fix 'copy to all' issue... Does it work / not work???
    -- We get called this when we need to 'set our schedule' via 'copy to all thermostats'...
    -- RYANE
    -- I do *NOT* care about this, since I'm not managing the events myself...
    return
  end

  if (strCommand == "SET_PRESETS") then
    -- Capture presets in global table...
    -- Parse <presets> XML for name, setpoints, preset fields...
    local xml = tParams["XML"] or ""
    if (xml == "") then return end

    local presets = XMLHelper.XMLToTable(xml)
    gPresets = {}

    for i = 1, #presets[1] do
      local mypreset = presets[1][i]
      local myargs = mypreset.xarg or {}
      local name = myargs["name"] or ""
      local pfxml = myargs["preset_fields"] or ""

      -- Must find both name and set of preset fields XML to proceed...
      if ((name ~= "") and (pfxml ~= "")) then
        gPresets[name] = {}

        -- Convert quoted XML back to XML...
        pfxml = pfxml:gsub("&gt;", ">"):gsub("&lt;", "<"):gsub("&quot;", '"'):gsub("&amp;", "&")

        dbg(pfxml)

        -- For each field in preset_fields, grab the id, value and store for later...
        local pvals = XMLHelper.XMLToTable(pfxml)
        for i = 1, #pvals[1] do
          local mypval = pvals[1][i]
          local mypargs = mypval.xarg or {}
          local id = mypargs["id"] or ""
          local value = mypargs["value"] or ""
          if (id ~= "") and (value ~= "") then
            gPresets[name][id] = value
          end
        end
      end
    end
    return
  end

  if (strCommand == "SET_PRESET") or (strCommand == "SET_EVENT") then
    local preset = tParams["NAME"] or tParams["PRESET"] or ""
    if (preset == "") then return end

    -- Hold Modes only effect Scheduled Event Preset changes, not SET_PRESET changes...
    if (strCommand == "SET_EVENT") then
      gLastScheduledPreset = preset -- Save Scheduled Presets to 'restore'...

      if (gHoldMode == "Permanent" or gHoldMode == "2 Hours") then
        dbg("Hold Mode " .. gHoldMode .. " prevents schedule change.")
        return
      end
      if (gHoldMode == "Until Next") then
        dbg("Hold Mode 'Until Next' ended.")
        gHoldMode = "Off"
        MySendToProxy(PROXY_BINDING, "HOLD_MODE_CHANGED", {MODE = gHoldMode})
      end
    end

    MySendToProxy(PROXY_BINDING, "PRESET_CHANGED", {NAME = preset})

    gPresets = gPresets or {}
    local mypreset = gPresets[preset] or ""
    if (mypreset == "") then return end

    if (mypreset.hvac_mode ~= nil) then
      ReceivedFromProxy(PROXY_BINDING, "SET_MODE_HVAC", {MODE = mypreset.hvac_mode, PRESET = true})
    end

    local tempval = mypreset.single_setpoint_c or ""
    if (tempval ~= "") then
      OneShotTimer.Add(2, "SECONDS",
        function()
          ReceivedFromProxy(PROXY_BINDING, "SET_SETPOINT_SINGLE", {CELSIUS = tempval, PRESET = true})
        end
        , "SPMC")
    end
    return
  end
end


function SetHeatCoolStatus()
  gHeatStatus = gHeatStatus or 0
  gCoolStatus = gCoolStatus or 0

  -- If both, show 'heating'...
  if (gHeatStatus > 0) then
    MySendToProxy(PROXY_BINDING, "HVAC_STATE_CHANGED", {STATE = "Heat"})
  elseif (gCoolStatus > 0) then
    MySendToProxy(PROXY_BINDING, "HVAC_STATE_CHANGED", {STATE = "Cool"})
  else
    MySendToProxy(PROXY_BINDING, "HVAC_STATE_CHANGED", {STATE = "Off"})
  end

end


FILTER = {}
FILTER["ANA"] = 1
FILTER["V1 "] = 1
FILTER["HOLD"] = 1
FILTER["HEAT"] = 1
FILTER["COOL"] = 1
FILTER["DR"] = 1
FILTER["REVERSING"] = 1
FILTER["SCHEDULE"] = 1
FILTER["DEADBAND"] = 1
FILTER["CALIBRATION"] = 1
FILTER["OUTDOOR"] = 1
FILTER["DISPLAY"] = 1
FILTER["VACATION"] = 1
FILTER["HUMIDITY_STATE"] = 1
FILTER["HUMIDITY_MODE"] = 1
FILTER["HUMIDIFY"] = 1


function ListDeviceVariables()
  local out, tins = {}, table.insert
  local proxyDeviceID = next(C4:GetBoundConsumerDevices(C4:GetDeviceID(), PROXY_BINDING))
  tins(out, "------------------- Proxy Variables -------------------")
  for k,v in pairs(C4:GetDeviceVariables(proxyDeviceID)) do
    local ok = true
    for filter in pairs(FILTER) do
      if (v.name:find(filter)) then ok = false end
    end
    if (v.name == "TEMPERATURE") then ok = false end
    if (ok) then tins(out, string.format("%-6d", k) .. v.name .. ": " .. v.value) end
  end
  tins(out, "------------------- Protocol Variables -------------------")
  for k,v in pairs(C4:GetDeviceVariables(C4:GetDeviceID())) do
    tins(out, string.format("%-6d", k) .. v.name .. ": " .. v.value)
  end
  print(table.concat(out, "\n"))
end


function ExecuteCommand(strCommand, tParams)
  tParams = tParams or {}
  strCommand = strCommand or ""
  tParams = tParams or {}
  if (strCommand == "") then return end

  --if (strCommand ~= "LASTCONNECTION_CHANGED") then
  if (not IGNORE_MESSAGES[strCommand]) then
    dbg("[<---- ExecC]: " .. strCommand .. " {" .. formatParams(tParams) .. "}")
  end

  if (strCommand == "OFFLINE") then
    SetConnected(false)
  end

  if (strCommand == "ONLINE") then
    SetConnected(true)
  end

  -- Handle KNX Feedback...
  if (strCommand == "DATA_FROM_KNX") then
    local ga = trim(tParams["GROUP_ADDRESS"])
    local value = tParams["VALUE"] or ""
    if (ga == "") or (value == "") then return end

    -- Find out which Property corresponds to this GA...
    for prop,propval in pairs(Properties) do
      propval = trim(propval)
      if (ga == propval) then
        if (prop == "Setpoint Status Address") then
          OneShotTimer.Add(50, "MILLISECONDS", function() MySendToProxy(PROXY_BINDING, "SINGLE_SETPOINT_CHANGED", {SCALE = "C", SETPOINT = tostring(value)}) end, "SSC")
          return
        end
        if (prop == "Temperature Status Address") then
          SetConnected(true)
          MySendToProxy(PROXY_BINDING, "TEMPERATURE_CHANGED", {SCALE = "C", TEMPERATURE = value})
          return
        end
        if (prop == "Mode Status Address") then
          OneShotTimer.Add(50, "MILLISECONDS", function() 
              MySendToProxy(PROXY_BINDING, "HVAC_MODE_CHANGED", {MODE = _gModeList[value]})
              SetRanges(_gModeList[value])
	    end
	    , "HMC")
	    return
        end
        if (prop == "Ventilation Status Address") then
          OneShotTimer.Add(50, "MILLISECONDS", function() 
              MySendToProxy(PROXY_BINDING, "FAN_MODE_CHANGED", {MODE = _gFanModeList[value]})
	    end
	    , "FMC")
	    return
        end
        if (prop == "Heating Status Address") then
	  gHeatStatus = tonumber_loc(value) or 0
	  SetHeatCoolStatus()
	  return
	end
        if (prop == "Cooling Status Address") then
	  gCoolStatus = tonumber_loc(value) or 0
	  SetHeatCoolStatus()
	  return
	end
      end
    end
  end

  if (strCommand == "FAN_MODE_CHANGED") then
    local fanmode = tParams["MODE"] or ""
    if (fanmode == "") then return end

    fanmode = trim(string.upper(fanmode))
    if (fanmode == "ON") then
      MySendToProxy(PROXY_BINDING, "ALLOWED_FAN_MODES_CHANGED", {MODES = "Off,Fan Timer"})
      MySendToProxy(PROXY_BINDING, "FAN_MODE_CHANGED", {MODE = "Fan Timer"})
      MySendToProxy(PROXY_BINDING, "FAN_STATE_CHANGED", {STATE = "On"})
    else
      MySendToProxy(PROXY_BINDING, "ALLOWED_FAN_MODES_CHANGED", {MODES = "Off,Fan Timer"})
      MySendToProxy(PROXY_BINDING, "FAN_MODE_CHANGED", {MODE = "Off"})
      MySendToProxy(PROXY_BINDING, "FAN_STATE_CHANGED", {STATE = "Off"})
    end
  end

  if (strCommand == "LUA_ACTION") then
    tParams = tParams or {}
    local action = tParams.ACTION or ""
    if (action == "ListVariables") then
      ListDeviceVariables()
    end
  end

  MySendToProxy(PROXY_BINDING, strCommand, tParams)
end


-- TODO: Property to turn on/off messages???? -- Also, message list array, command to set/clear messages...
function MySetMessage(strMsg)
  MySendToProxy(PROXY_BINDING, "MESSAGE_CHANGED", {MESSAGE = strMsg})
end


TO_BINDING = {}
TO_BINDING[KNX_BINDING] = "----> KNX  "
TO_BINDING[PROXY_BINDING] = "----> Proxy"

IGNORE_MESSAGES = {}
IGNORE_MESSAGES["EXTRAS_STATE_CHANGED"] = false
IGNORE_MESSAGES["UPDATE_SCHEDULE_ENTRIES"] = true


function MySendToProxy(idBinding, strCommand, tParams)

  if (not IGNORE_MESSAGES[strCommand]) then
    dbg("[" .. TO_BINDING[idBinding] .. "]: " .. strCommand .. " {" .. formatParams(tParams) .. "}")
  end

  C4:SendToProxy(idBinding, strCommand, tParams)
end


function RealInit()
  gModeList, _gModeList = ParseModeList()
  gFanModeList, _gFanModeList = ParseFanModeList()
  gModeRange = ParseModeRanges()

  OnPropertyChanged("Setpoint Resolution")
  OnPropertyChanged("Temperature Resolution")
  OnPropertyChanged("Hold Modes")
  OnPropertyChanged("Ventilation Mode List")
  OnPropertyChanged("Ventilation Control")

  -- Restore Preset fields to last setting, so we don't lose preset on restart.
  RestorePresetScheduling()

end


function OnDriverLateInit()
  PersistData = PersistData or {}
  PersistData.Scale = PersistData.Scale or "C"
  if (PersistData.Scale == "C") or (PersistData.Scale == "CELSIUS") then
    PersistData.Scale = "CELSIUS"
  else
    PersistData.Scale = "FAHRENHEIT"
  end

  MySendToProxy(PROXY_BINDING, "SCALE_CHANGED", {SCALE = PersistData.Scale})
  SetConnected(false)

  -- Removed Extras setups, no Preset Scheduling change, Leaf Mode, or Pincode, Away Status in Extras. Kept above as an example.
  MySendToProxy(PROXY_BINDING, "EXTRAS_SETUP_CHANGED", {XML = [[
      <extras_setup>
      <extra>
      </extra>
      </extras_setup>
      ]]})

  --MySendToProxy(PROXY_BINDING, "EXTRAS_STATE_CHANGED", {XML = '<extras_state><extra><object id="leaf" value="False"/></extra></extras_state>'})

  OneShotTimer.Add(5, "SECONDS", RealInit, "REAL_LATE_INIT")

  -- Clear any messages...
  MySetMessage("")

  MySendToProxy(PROXY_BINDING, "ALLOWED_HOLD_MODES_CHANGED", {MODES = " "})
  MySendToProxy(PROXY_BINDING, "FAN_MODE_CHANGED", {MODE = "Off"})

  MySendToProxy(PROXY_BINDING, "HVAC_STATE_CHANGED", {STATE = "Off"})
  MySendToProxy(PROXY_BINDING, "FAN_STATE_CHANGED", {STATE = "Off"})

  -- By default, start with no humidity tab... If HUMIDITY_CHANGED received, make tab visible and send humidity value...
  PersistData = PersistData or {}
  local humidity = PersistData.Humidity or false
  MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {HAS_HUMIDITY = humidity})
  MySendToProxy(PROXY_BINDING, "HUMIDITY_CHANGED", {HUMIDITY = 0})

  OneShotTimer.Add(10, "SECONDS", AddGroupItems, "AGI")
end


function SetConnected(bConnected)
  gConnected = gConnected or ""
  if (bConnected ~= gConnected) then
    MySendToProxy(PROXY_BINDING, "CONNECTION", {CONNECTED = bConnected})
    if (bConnected) then
      MySetMessage("")
    else
      MySetMessage("KNX Offline")
    end
  end
  gConnected = bConnected
end


print("Driver Loaded: " .. os.date())


----------------------- INCLUDES ---------------------------
-- Now allows for 'named' one-shot timers.  If the name is re-used, reset timer, don't make a new one...

---------------------------------------------------------
--------------  OneShotTimer Timers Module  ----------------
---------------------------------------------------------
do
  local OTE, ODD
  local MTFUNC = function() end

  OneShotTimer = {}
  local tCallback = {}
  
  function OneShotTimer.OnTimerExpired(idTimer)
    if (tCallback[idTimer] ~= nil) then
      C4:KillTimer(idTimer)
      tCallback[idTimer].CALLBACK(tostring(tCallback[idTimer].NAME))
      tCallback[idTimer] = nil
    else
      OTE(idTimer)
    end
  end

  function OneShotTimer.ClearAll()
    for k,v in pairs(tCallback) do C4:KillTimer(k) end
    tCallback = {}
  end

  function OneShotTimer.OnDriverDestroyed()
    OneShotTimer.ClearAll()
    ODD()
  end

  function OneShotTimer.Kill(Name)
    if (Name ~= nil) then
      for k,v in pairs(tCallback) do if (v.NAME == Name) then C4:KillTimer(k) tCallback[k] = nil end end
    end
  end

  function OneShotTimer.Add(nInterval, strUnits, fCallback, Name)
    local id = C4:AddTimer(nInterval, strUnits)
    --print("Adding OneShotTimer[" .. nInterval .. " " .. strUnits .. "]: " .. Name)

    -- Look for name if not nil, if found, remove existing timer callback...
    if (Name ~= nil) then
      for k,v in pairs(tCallback) do
        if (v.NAME == Name) then 
	  --print("Killing OneShotTimer Callback: " .. Name)
          C4:KillTimer(k)
          tCallback[k] = nil
        end
      end
    end
    tCallback[id] = {CALLBACK = fCallback, NAME = Name}
  end

  -----  Overrides:  --------
  OTE , OnTimerExpired = OnTimerExpired or MTFUNC, OneShotTimer.OnTimerExpired
  ODD , OnDriverDestroyed = OnDriverDestroyed or MTFUNC, OneShotTimer.OnDriverDestroyed
  ---  END Overrides:  ------
end
---------------------------------------------------------
-----------  END OneShotTimer Timers Module  ---------------
---------------------------------------------------------

-- TODO: Make this a true module, by returning it inside a local table...

------------------------------------------------------
-----------------  XML Helper Module   ------------------
------------------------------------------------------

-- Externs:
-- XMLToTable(strXML))
-- XMLPath(tXML, strWalk)
-- TableToXML(tXML)
  -- Examples:
  -- {params = {param = { name = "MyValue", value = {static = {type = {INTEGER = 123}}}}}}
  -- {http_api = {session = {device_id = 1, tpnumber = 3}}}
-- MakeXML(strTag, strValue, tAttrs)
  -- Examples:
  -- MakeXML("c4soap", table.concat(params, ''), {name = strCommand, seq = mSeq, async = TF[bAsync]})
-- XMLSpecialChars

do
  XMLHelper = {}

  local XMLCHARS = {
    ["&apos;"] = "'",
    ["&lt;"] = "<",
    ["&gt;"] = ">",
    ["&quot;"] = "'",
    ["&amp;"] = "&",
  }


  function XMLHelper.XMLSpecialChars(name)
    local s_gsub = string.gsub
    for k,v in pairs(XMLCHARS) do name = s_gsub(name, k, v) end
    return name
  end


  ---- SIMPLE XML PARSER from http://lua-users.org/wiki/LuaXml (parseargs and collect originally)------------
  local function parseargs(s)
    local arg = {}
    string.gsub(s, "([%w_]+)=([\"'])(.-)%2", function (w, _, a)
      arg[w] = a
    end)
    return arg
  end
    
  function XMLHelper.XMLToTable(s)
    local stack = {}
    local top = {}
    table.insert(stack, top)
    local ni,c,label,xarg, empty
    local i, j = 1, 1
    while true do
      ni,j,c,label,xarg, empty = string.find(s, "<(%/?)([%w_]+)(.-)(%/?)>", i)
      if not ni then break end
      local text = string.sub(s, i, ni-1)
      if not string.find(text, "^%s*$") then
        table.insert(top, text)
      end
      if empty == "/" then  -- empty element tag
        table.insert(top, {label=label, xarg=parseargs(xarg), empty=1})
      elseif c == "" then   -- start tag
        top = {label=label, xarg=parseargs(xarg)}
        table.insert(stack, top)   -- new level
      else  -- end tag
        local toclose = table.remove(stack)  -- remove top
        top = stack[#stack]
        if #stack < 1 then
          error("nothing to close with "..label)
        end
        if toclose.label ~= label then
          error("trying to close "..toclose.label.." with "..label)
        end
        table.insert(top, toclose)
      end
      i = j+1
    end
    local text = string.sub(s, i)
    if not string.find(text, "^%s*$") then
      table.insert(stack[#stack], text)
    end
    if #stack > 1 then
      error("unclosed "..stack[#stack].label)
    end
    return stack[1]
  end


  -- Converts tables with sub-tables into <key><subkey>subvalue</subkey></key> XML...
  -- NOTE: Does *not* support attributes...
  function XMLHelper.TableToXML(tXML)
    if (type(tXML) ~= "table") then return tostring(tXML) end
    local val = ""

    local k,v = next(tXML)
    while (k ~= nil) do
      val = val .. XMLHelper.MakeXML(k, XMLHelper.TableToXML(v))
      k,v = next(tXML, k)
    end
    return val
  end


  function XMLHelper.MakeXML(strTag, strValue, tAttrs)
    local xml, p = {}, {}
    for k,v in pairs(tAttrs or {}) do
      table.insert(p, ' ' .. k .. '="' .. v .. '"')
    end
    table.insert(xml, '<' .. strTag .. table.concat(p, '') .. '>' .. strValue .. '</' .. strTag .. '>')
    return table.concat(xml, '')
  end


  function XMLHelper.XMLPath(tXML, strWalk)
    local out = {}
    -- Do a sort-of-like XPath walk through the Tableified XML...
    -- Examples:
    -- XMLHelper.XMLPath(t, "^c4soap")  -- returns inner *text* of <c4soap attr="Value1">InnerText</c4soap>
    -- XMLHelper.XMLPath(t, "c4soap/physicaldevices/*physicaldevice")  -- returns all sub-items (<physicaldevice> under <c4soap><physicaldevices></></>)
        -- For each <physicaldevice, the attributes of the uuid/address/etc. are in (k,v) table, v.xarg.uuid/v.xarg.address/etc.
    -- XMLHelper.XMLPath(t, "#attr")  -- returns attribute named 'attr' of <c4soap attr="Value1">InnerText</c4soap>
    -- NOTE: Requires that a *table* be passed in, not the actual XML...  XMLHelper.XMLToTable is expected...

    if (strWalk == nil) then return nil end

    local pos = string.find(strWalk, "([/])")
    if (pos ~= nil) then 
      look = string.sub(strWalk, 1, pos - 1)
      walk = string.sub(strWalk, pos + 1)
    else
      look = strWalk
      walk = nil
    end

    local stype = string.sub(look, 1, 1)
    if ((stype == '^') or (stype == '#') or (stype == '*')) then
      look = string.sub(look, 2)
    else
      stype = ""
    end

    if (stype == '#') then
      if (type(tXML.xarg[look]) == "string") then
        return tXML.xarg[look]
      else
        return nil
      end
    end

    local i = 1
    while (tXML[i] ~= nil) do
      if (tXML[i].label == look) then
        if (stype == '*') then
          table.insert(out, tXML[i])
        elseif (stype == '^') then
          if (type(tXML[i][1]) == "string") then
            return tXML[i][1]
          else
            return nil
          end
        else
          -- stype == "" -- next substring...
          return(XMLHelper.XMLPath(tXML[i], walk))
        end
      end
      i = i + 1
    end

    if (#out > 0) then 
      return out 
    else 
      return nil 
    end
  end

end
----------------------------------------------------------
-----------------  END XML Helper Module   ------------------
----------------------------------------------------------

------------------------------------------------------------

-----------------------------------------------------
------------------------ INIT ------------------------
-----------------------------------------------------
gHoldMode = "Off"
if (LuaC4Object.AllowExecute) then C4:AllowExecute(true) end

OneShotTimer.Add(15, "SECONDS", RealInit, "REAL_LATE_INIT")

